//! Web frontend for the crates registry.
//!
//! The frontend is minimal, basic HTML, generated from Handlebars
//! templates.
//!
//! Rocket looks for templates by default in the `./templates`
//! directory relative to the crate root. This is convenient for
//! development. If the `bundle-templates` feature is enabled, then
//! the templates will be bundled into the binary and unpacked into a
//! tempdir.

use crate::{auth, database::Database};
use rocket::{Route, State};
use rocket_dyn_templates::Template;
use serde::Serialize;
use std::collections::BTreeMap;
use tracing::warn;

/// Unpack template files into a temporary directory. The `TempDir`
/// must not be dropped in order for the templates to remain
/// available.
#[cfg(feature = "bundle-templates")]
pub async fn unpack() -> Result<tempfile::TempDir, std::io::Error> {
    static TEMPLATES: include_dir::Dir = include_dir::include_dir!("./templates");

    let template_dir = tempfile::Builder::new()
        .prefix("crates-templates")
        .tempdir()?;

    // Skip the first entry, because it is the root directory, which
    // already exists.
    for entry in TEMPLATES.find("**/*").unwrap().skip(1) {
        match entry {
            include_dir::DirEntry::Dir(dir) => {
                tokio::fs::create_dir(template_dir.path().join(dir.path())).await?;
            }
            include_dir::DirEntry::File(file) => {
                tokio::fs::write(template_dir.path().join(file.path()), file.contents()).await?;
            }
        }
    }

    Ok(template_dir)
}

#[derive(Serialize)]
struct Context {
    title: &'static str,
    parent: &'static str,
}

pub fn routes() -> Vec<Route> {
    rocket::routes![index, login, login_code, login_provider]
}

#[rocket::get("/")]
fn index() -> Template {
    Template::render(
        "index",
        Context {
            title: "Home",
            parent: "layout",
        },
    )
}

#[derive(Serialize)]
struct LoginContext {
    title: &'static str,
    providers: Vec<LoginProvider>,
    parent: &'static str,
}

#[derive(Serialize)]
struct LoginProvider {
    name: String,
    url: String,
}

#[rocket::get("/login")]
fn login(providers: &State<BTreeMap<String, auth::Provider>>) -> Template {
    let providers = providers
        .iter()
        .map(|(provider_id, provider)| LoginProvider {
            name: provider.name.clone(),
            url: format!("/login/{}", provider_id),
        })
        .collect();

    Template::render(
        "login",
        LoginContext {
            title: "Login",
            providers,
            parent: "layout",
        },
    )
}

#[derive(Serialize)]
struct ApiKeyContext {
    title: &'static str,
    token: String,
    parent: &'static str,
}

#[rocket::get("/login?<code>&<state>")]
async fn login_code(
    code: String,
    state: String,
    providers: &State<BTreeMap<String, auth::Provider>>,
    database: &State<Database>,
) -> Template {
    let token = async {
        let provider_id = state
            .split(":")
            .next()
            .expect("str::split is guaranteed to yield at least one element");

        if let Some(provider) = providers.get(provider_id) {
            match crate::auth::create_token(code, provider_id, provider, &database).await {
                Err(err) => {
                    warn!(%err, "Failed to create auth token");
                    None
                }
                Ok(token) => Some(token),
            }
        } else {
            warn!(%state, "Couldn't find OAuth provider");
            None
        }
    };

    match token.await {
        Some(token) => Template::render(
            "api_token",
            ApiKeyContext {
                title: "Login",
                token,
                parent: "layout",
            },
        ),
        None => login(providers),
    }
}

#[rocket::get("/login/<provider_id>")]
fn login_provider(
    provider_id: String,
    providers: &State<BTreeMap<String, auth::Provider>>,
) -> rocket::response::Redirect {
    let provider = match providers.get(&provider_id) {
        Some(provider) => provider,
        None => return rocket::response::Redirect::to(rocket::uri!(login)),
    };

    let state = format!("{}:{}", provider_id, token(8));

    let mut url = provider.url.join("authorize").unwrap();
    url.query_pairs_mut()
        .append_pair("client_id", &provider.client_id)
        .append_pair("redirect_uri", &provider.redirect_url)
        .append_pair("scope", &provider.scope)
        .append_pair("response_type", "code")
        .append_pair("state", &state);
    rocket::response::Redirect::to(url.to_string())
}

fn token(len: usize) -> String {
    use rand::Rng;
    rand::thread_rng()
        .sample_iter(&rand::distributions::Alphanumeric)
        .take(len)
        .map(char::from)
        .collect()
}
