//! Metrics for Prometheus.
//!
//! This module defines various metrics and exposes a [`Fairing`] that
//! tracks metrics and attaches a `/metrics` endpoint. Attach it using
//! the [`Rocket::attach()`] function:
//!
//! ```
//! rocket::ignite().attach(Metrics)
//! ```

use prometheus::*;
use prometheus_static_metric::*;
use rocket::{
    fairing::{self, Fairing, Info, Kind},
    http::Method,
    Build, Data, Request, Rocket,
};

make_static_metric! {
    pub struct HttpRequestDuration: Histogram {
        "method" => { get, put, delete, other }
    }
}

lazy_static::lazy_static! {
    /// The amount of time spent handling HTTP requests.
    pub static ref HTTP_REQUEST_DURATION: HttpRequestDuration = register_static_histogram_vec!(
        HttpRequestDuration,
        "http_request_duration",
        "Duration of time spent processing requests.",
        &["method"],
    ).unwrap();
}

/// A Rocket [`Fairing`] that tracks metrics and exposes a `/metrics`
/// endpoint for Prometheus.
pub struct Metrics;

#[rocket::async_trait]
impl Fairing for Metrics {
    fn info(&self) -> Info {
        Info {
            name: "Metrics",
            kind: Kind::Ignite | Kind::Request,
        }
    }

    async fn on_ignite(&self, rocket: Rocket<Build>) -> fairing::Result {
        Ok(rocket.mount("/", rocket::routes![endpoint]))
    }

    async fn on_request(&self, request: &mut Request<'_>, _: &mut Data<'_>) {
        let histogram = match request.method() {
            Method::Get => &HTTP_REQUEST_DURATION.get,
            Method::Put => &HTTP_REQUEST_DURATION.put,
            Method::Delete => &HTTP_REQUEST_DURATION.delete,
            _ => &HTTP_REQUEST_DURATION.other,
        };
        request.local_cache(|| histogram.start_timer());
    }
}

#[rocket::get("/metrics")]
fn endpoint() -> String {
    use prometheus::Encoder;
    let mut buffer = Vec::new();
    let encoder = prometheus::TextEncoder::new();
    let metric_families = prometheus::gather();
    encoder.encode(&metric_families, &mut buffer).unwrap();
    String::from_utf8(buffer).unwrap()
}
